from .parsing_enum import ParsingEnum


class Breed(ParsingEnum):
    """
    An enumeration representing a breed of horse.
    """

    THOROUGHBRED = 1
    ARABIAN = 2
    QUARTER_HORSE = 3
    AQPS = 4

    # Abbreviations
    TB = THOROUGHBRED
    QH = QUARTER_HORSE



import re
from decimal import Decimal
from measurement.measures import Distance  # type: ignore


class HorseHeight(Distance):
    """
    A class for measuring a horse's height, created by wrapping the measurement library Distance class to allow
    for the use of hands as a unit.
    """

    def __init__(self, distance: str) -> None:
        """
        Initialize a HorseHeight object from a string.
        """
        pattern = re.compile(r"(\d+\D+)")
        vals_and_units = pattern.findall(distance.replace(" ", ""))

        height = Distance(inch=0)
        for vu in vals_and_units:
            matches = re.compile(r"(\d+)(\D+)").match(vu)
            if matches:
                val, unit = matches.groups()
                if unit == "hands" or unit == "hh":
                    height += Distance(inch=int(val) * 4)
                else:
                    height += Distance(**{unit: int(val)})

        super().__init__(self, inch=height.inch)

    @property
    def hand(self) -> Decimal:
        """
        Returns the height in hands.
        """
        return Decimal(self.inch / 4)



from typing import Self
import pendulum
from pendulum import DateTime, Period


class HorseAge:
    def __init__(
        self,
        official_age: int | None = None,
        *,
        foaling_date: DateTime | None = None,
        birth_year: int | None = None,
        context_date: DateTime | None = None
    ):
        if official_age is not None and (foaling_date or birth_year):
            raise ValueError(
                "Cannot initialize HorseAge with both official_age and keyword"
            )

        if not (official_age or foaling_date or birth_year):
            raise ValueError(
                "Cannot initialize HorseAge without official_age, foaling_date, or birth_year"
            )

        self._actual_dob = foaling_date
        self._context_date = context_date

        year = (
            birth_year
            if birth_year
            else self._base_date.year - official_age
            if official_age
            else foaling_date.year
            if foaling_date
            else None
        )

        self._official_dob = pendulum.datetime(year, 1, 1) if year else None

    @property
    def official(self) -> Period:
        if not self._official_dob:
            raise ValueError("Cannot calculate official age as official dob is unknown")
        return self._calc_age(self._official_dob)

    @property
    def actual(self) -> Period:
        if not self._actual_dob:
            raise ValueError("Cannot calculate actual age as actual dob is unknown")
        return self._calc_age(self._actual_dob)

    @property
    def _base_date(self) -> DateTime:
        return self._context_date if self._context_date else pendulum.now()

    def at(self, date: DateTime) -> Self:
        self._context_date = date
        return self

    def _calc_age(self, dob: DateTime) -> Period:
        return (
            self._base_date - dob
            if dob < self._base_date
            else self._base_date - self._base_date
        )



from .parsing_enum import ParsingEnum


class Headgear(ParsingEnum):
    """
    An enumeration representing the headgear worn by a horse in a race.
    """

    HOOD = 1
    BLINKERS = 2
    CHEEKPIECES = 3
    TONGUE_TIE = 4
    VISOR = 5
    EYE_HOOD = 6
    EYE_COVER = 7
    EYE_SHIELD = 8

    # Abbreviations
    H = HOOD
    B = BLINKERS
    BL = BLINKERS
    C = CHEEKPIECES
    P = CHEEKPIECES
    T = TONGUE_TIE
    TT = TONGUE_TIE
    V = VISOR
    E = EYE_HOOD
    EC = EYE_COVER
    ES = EYE_SHIELD



from decimal import Decimal, InvalidOperation
from enum import Enum
from fractions import Fraction
from typing import Optional, Self


class Horselength(Decimal):
    """
    A Decimal subclass that represents a distance in horse lengths.

    Horse lengths are commonly used to measure the distance between horses in a race.
    """

    class Description(Enum):
        """
        An enumeration that represents some commonly used lengths in horse racing,
        with their corresponding values.

        Abbreviations can also be used to reference these lengths.
        """

        DEAD_HEAT = 0
        NOSE = 0.05
        SHORT_HEAD = 0.1
        HEAD = 0.2
        SHORT_NECK = 0.25
        NECK = 0.3
        DISTANCE = 100

        # Abbreviations
        DHT = DEAD_HEAT
        NS = NOSE
        NSE = NOSE
        SHD = SHORT_HEAD
        HD = HEAD
        SNK = SHORT_NECK
        NK = NECK
        DIST = DISTANCE
        DST = DISTANCE

    def __new__(cls, value: Optional[int | float | Decimal | str] = None) -> Self:
        """Create a new horselength instance

        :param value: Either a description or a number, representing the horselength, defaults to None
        :type value: Optional[int  |  float  |  Decimal  |  str], optional
        :raises ValueError: For invalid inputs
        :return: A new horselength instance
        :rtype: Self
        """
        value = str(value or 0)
        if "/" in value:
            parts = value.split(" ")
            whole = parts[0] if len(parts) == 2 else 0
            fraction = Fraction(parts[-1])
            value = Decimal(str(whole)) + Decimal(fraction.numerator) / Decimal(
                fraction.denominator
            )
        elif value.upper() in cls.Description.__members__:
            value = str(cls.Description[value.upper()].value)

        try:
            return super().__new__(cls, value)
        except InvalidOperation:
            raise ValueError(f"Invalid input value for Horselength: {value}")

    def __repr__(self) -> str:
        """
        Return a representation of the Horselength instance.
        """
        return f"Horselength('{self}')"

    def __str__(self) -> str:
        """
        Return a string representation of the Horselength instance.
        """
        if self % 1 == 0:
            return super().__str__()
        else:
            for e in Horselength.Description:
                if self == Decimal(str(e.value)):
                    return e.name.lower()

            whole = self // 1
            fraction = Fraction(str(self % 1))
            return f"{whole} {fraction}" if whole else str(fraction)

    def __add__(self, other: Decimal | int) -> "Horselength":
        """
        Adds two Horselength instances or an instance and a Decimal and returns a Horselength.
        """
        sum = super().__add__(other)
        return Horselength(sum)

    def __radd__(self, other: Decimal | int) -> "Horselength":
        """
        Adds two Horselength instances or an instance and a Decimal and returns a Horselength.
        """
        sum = super().__radd__(other)
        return Horselength(sum)

    def __sub__(self, other: Decimal | int) -> "Horselength":
        """
        Subtracts two Horselength instances or an instance and a Decimal and returns a Horselength.
        """
        diff = super().__sub__(other)
        return Horselength(diff)

    def __rsub__(self, other: Decimal | int) -> "Horselength":
        """
        Subtracts two Horselength instances or an instance and a Decimal and returns a Horselength.
        """
        diff = super().__rsub__(other)
        return Horselength(diff)



import re
from decimal import Decimal
from measurement.measures import Distance  # type: ignore


class RaceDistance(Distance):
    """
    A thin wrapper around the measurement library Distance class to allow for the creation of Distance objects
    from strings and to provide a way to initialize with furlongs.
    """

    def __init__(self, distance: str) -> None:
        """
        Initialize a RaceDistance object from a string.
        """
        pattern = re.compile(r"(\d+\D+)")
        unit_dict = {"m": "mile", "f": "furlong", "y": "yard"}
        vals_and_units = pattern.findall(distance.replace(" ", ""))

        distance = Distance(m=0)
        for vu in vals_and_units:
            matches = re.compile(r"(\d+)(\D+)").match(vu)
            if matches:
                val, unit = matches.groups()
                unit = unit_dict[unit] if unit in unit_dict and int(val) < 221 else unit
                if unit == "furlong":
                    distance += Distance(chain=int(val) * 10)
                else:
                    distance += Distance(**{unit: int(val)})

        super().__init__(self, m=distance.m)  # type: ignore

    @property
    def furlong(self) -> Decimal:
        """
        Returns the distance in furlongs.
        """
        return Decimal(self.chain / 10)



from .parsing_enum import ParsingEnum


class Gender(ParsingEnum):
    """
    An enumeration representing the gender of a horse
    """

    FOAL = 0
    YEARLING = 1
    COLT = 2
    FILLY = 3
    STALLION = 4
    MARE = 5
    GELDING = 6
    RIG = 7

    # Abbreviations
    C = COLT
    F = FILLY
    S = STALLION
    M = MARE
    G = GELDING
    R = RIG

    # Plural
    FOALS = FOAL
    YEARLINGS = YEARLING
    COLTS = COLT
    FILLIES = FILLY
    STALLIONS = STALLION
    MARES = MARE
    GELDINGS = GELDING
    RIGS = RIG



from .parsing_enum import ParsingEnum


class RacingCode(ParsingEnum):
    FLAT = 1
    NATIONAL_HUNT = 2
    POINT_TO_POINT = 3

    # Abbreviations
    F = FLAT
    N = NATIONAL_HUNT
    NH = NATIONAL_HUNT
    P = POINT_TO_POINT
    PTP = POINT_TO_POINT
    P2P = POINT_TO_POINT



from enum import Enum, EnumMeta


class ParsingEnumMeta(EnumMeta):
    def __getitem__(cls, name):
        for member in cls.__members__:
            test_val = name.replace("-", " ").replace(" ", "_").replace("'", "").upper()
            if member == test_val:
                return cls.__members__[member]
        raise KeyError(name)


class ParsingEnum(Enum, metaclass=ParsingEnumMeta):
    pass



from .parsing_enum import ParsingEnum


class Disaster(ParsingEnum):
    """
    An enumeration representing the reason for a horse's non-completion of a race
    """

    FELL = 1
    REFUSED = 2
    BROUGHT_DOWN = 3
    UNSEATED_RIDER = 4
    PULLED_UP = 5
    SLIPPED_UP = 6
    CARRIED_OUT = 7
    RAN_OUT = 8
    LEFT_AT_START = 9
    HIT_RAIL = 10
    DISQUALIFIED = 11

    # Abbreviations
    F = FELL
    R = REFUSED
    REF = REFUSED
    B = BROUGHT_DOWN
    BD = BROUGHT_DOWN
    U = UNSEATED_RIDER
    UR = UNSEATED_RIDER
    P = PULLED_UP
    PU = PULLED_UP
    S = SLIPPED_UP
    SU = SLIPPED_UP
    C = CARRIED_OUT
    CO = CARRIED_OUT
    O = RAN_OUT
    RO = RAN_OUT
    L = LEFT_AT_START
    LEFT = LEFT_AT_START
    HR = HIT_RAIL
    D = DISQUALIFIED
    DQ = DISQUALIFIED
    DSQ = DISQUALIFIED



from .parsing_enum import ParsingEnum


class CoatColour(ParsingEnum):
    """
    An enumeration representing a coat colour of a horse.
    """

    BAY = 1
    BLACK = 2
    CHESTNUT = 3
    DARK_BAY = 4
    GREY = 5
    PALOMINO = 6
    WHITE = 7
    ROAN = 8

    # Abbreviations
    B = BAY
    BL = BLACK
    C = CHESTNUT
    CH = CHESTNUT
    DB = DARK_BAY
    BR = DARK_BAY
    G = GREY
    GR = GREY
    GRAY = GREY
    P = PALOMINO
    W = WHITE
    R = ROAN
    RO = ROAN



from .parsing_enum import ParsingEnum


class DirtGoingDescription(ParsingEnum):
    """
    An enumeration that represents a scale of US dirt going descriptions.
    """

    SEALED = 0
    SLOW = 1
    SLOPPY = 2
    MUDDY = 3
    GOOD = 4
    WET_FAST = 5
    FAST = 6



from .parsing_enum import ParsingEnum


class AWGoingDescription(ParsingEnum):
    """
    An enumeration that represents a scale of UK and Ireland all-weather going descriptions.

    """

    SLOW = 1
    STANDARD_TO_SLOW = 2
    STANDARD = 3
    STANDARD_TO_FAST = 4
    FAST = 5

    # Abbreviations
    SLW = SLOW
    STS = STANDARD_TO_SLOW
    STD = STANDARD
    STF = STANDARD_TO_FAST
    FST = FAST



from .age_category import AgeCategory
from .aw_going_description import AWGoingDescription
from .breed import Breed
from .coat_colour import CoatColour
from .dirt_going_description import DirtGoingDescription
from .disaster import Disaster
from .horse_experience_level import HorseExperienceLevel
from .jockey_experience_level import JockeyExperienceLevel
from .finishing_position import FinishingPosition
from .form_break import FormBreak
from .form_figures import FormFigures
from .gender import Gender
from .headgear import Headgear
from .horse_age import HorseAge
from .horse_height import HorseHeight
from .horselength import Horselength
from .obstacle import Obstacle
from .race_designation import RaceDesignation
from .race_distance import RaceDistance
from .race_title import RaceTitle
from .racing_code import RacingCode
from .sex import Sex
from .surface import Surface
from .turf_going_description import TurfGoingDescription

__all__ = [
    "AgeCategory",
    "AWGoingDescription",
    "Breed",
    "CoatColour",
    "DirtGoingDescription",
    "Disaster",
    "FinishingPosition",
    "FormBreak",
    "FormFigures",
    "Gender",
    "Headgear",
    "HorseAge",
    "HorseHeight",
    "Horselength",
    "HorseExperienceLevel",
    "JockeyExperienceLevel",
    "Obstacle",
    "RaceDesignation",
    "RaceDistance",
    "RaceExperienceStatus",
    "RaceTitle",
    "RacingCode",
    "Sex",
    "Surface",
    "TurfGoingDescription",
]



from .parsing_enum import ParsingEnum


class Sex(ParsingEnum):
    """
    An enumeration representing the sex of a horse
    """

    MALE = 1
    FEMALE = 2

    # Abbreviations
    M = MALE
    XY = MALE
    F = FEMALE
    XX = FEMALE



from horsetalk import Disaster, FinishingPosition, FormBreak


class FormFigures:
    @staticmethod
    def parse(form_figures: str):
        return [
            FinishingPosition(int(figure))
            if figure.isdigit()
            else FormBreak(figure)
            if figure in [member.value for member in FormBreak]
            else Disaster[figure]
            for figure in form_figures
        ]



from enum import Enum


class FinishingPosition(Enum):
    UNPLACED = 0
    FIRST = 1
    SECOND = 2
    THIRD = 3
    FOURTH = 4
    FIFTH = 5
    SIXTH = 6
    SEVENTH = 7
    EIGHTH = 8
    NINTH = 9

    def __str__(self):
        return str(self.value)



from .parsing_enum import ParsingEnum


class RaceDesignation(ParsingEnum):
    HANDICAP = 1
    CONDITIONS = 2
    AUCTION = 3
    CLAIMER = 4
    SELLER = 5

    # Abbreviations
    HCAP = HANDICAP
    AU = AUCTION
    CL = CLAIMER
    S = SELLER

    # Alternatives
    CLAIMING = CLAIMER
    SELLING = SELLER



from .parsing_enum import ParsingEnum


class JockeyExperienceLevel(ParsingEnum):
    AMATEUR = 1
    CONDITIONAL = 2
    APPRENTICE = 3
    PROFESSIONAL = 4

    # Plurals
    AMATEURS = AMATEUR
    CONDITIONALS = CONDITIONAL
    APPRENTICES = APPRENTICE
    PROFESSIONALS = PROFESSIONAL



from .parsing_enum import ParsingEnum


class Obstacle(ParsingEnum):
    """
    An enumeration representing the type of obstacle a race takes place over
    """

    HURDLE = 1
    STEEPLECHASE = 2
    CROSS_COUNTRY = 3

    # Abbreviations
    H = HURDLE
    HRD = HURDLE
    HRDLE = HURDLE
    C = STEEPLECHASE
    CH = STEEPLECHASE
    CHS = STEEPLECHASE
    CHSE = STEEPLECHASE
    CHASE = STEEPLECHASE
    CC = CROSS_COUNTRY
    XC = CROSS_COUNTRY



from .parsing_enum import ParsingEnum


class Surface(ParsingEnum):
    """
    An enumeration representing the surface upon which races are held
    """

    TURF = 1
    DIRT = 2
    ALL_WEATHER = 3

    # Abbreviations
    T = TURF
    D = DIRT
    AW = ALL_WEATHER



from enum import Enum
from typing import List, Type
from .age_category import AgeCategory
from .horse_experience_level import HorseExperienceLevel
from .gender import Gender
from .obstacle import Obstacle
from .race_designation import RaceDesignation


class RaceTitle:
    _words: List[str] = []

    @classmethod
    def parse(cls, title: str) -> dict:
        """Parse a race title into component parts

        :param title: A race title
        :type title: str
        :return: A dictionary of component parts
        :rtype: dict
        """
        self = cls()
        self._words = title.split()

        enums = [AgeCategory, HorseExperienceLevel, Gender, Obstacle, RaceDesignation]
        end_index = -1
        for i, word in enumerate(self._words):
            if any(getattr(enum, word, None) is not None for enum in enums):
                end_index = i
                break
        name = " ".join(self._words[:end_index])

        return {
            "age_category": self._lookup(AgeCategory),
            "horse_experience_level": self._lookup(HorseExperienceLevel),
            "gender": self._lookup(Gender, allow_multiple=True),
            "obstacle": self._lookup(Obstacle),
            "race_designation": self._lookup(RaceDesignation),
            "name": name,
        }

    def _lookup(
        self, enum: Type[Enum], allow_multiple: bool = False
    ) -> List[Enum] | Enum | None:
        """Private method to lookup an enum value from a list of words

        :param enum: Enum to search through
        :type enum: Type[Enum]
        :return: The found Enum value or None
        :rtype: Enum | None
        """
        found_values = [
            found_value
            for word in self._words
            if (found_value := getattr(enum, word, None)) is not None
        ]
        return (
            None
            if not found_values
            else found_values
            if allow_multiple
            else found_values[-1]
        )



from horsetalk.parsing_enum import ParsingEnum


class TurfGoingDescription(ParsingEnum):
    """
    An enumeration that represents a scale of UK or Ireland turf going descriptions.

    Values are rough equivalents to GoingStick readings.
    """

    HEAVY = 5
    SOFT = 6
    YIELDING_TO_SOFT = 6.5
    GOOD_TO_SOFT = 7
    YIELDING = 7
    GOOD_TO_YIELDING = 7.5
    GOOD = 8
    GOOD_TO_FIRM = 9
    FIRM = 10
    HARD = 11

    # Abbreviations
    V = HEAVY
    HV = HEAVY
    HVY = HEAVY
    S = SOFT
    SFT = SOFT
    D = GOOD_TO_SOFT
    GS = GOOD_TO_SOFT
    YS = YIELDING_TO_SOFT
    Y = YIELDING
    YLD = YIELDING
    GY = GOOD_TO_YIELDING
    G = GOOD
    GD = GOOD
    GF = GOOD_TO_FIRM
    M = GOOD_TO_FIRM
    F = FIRM
    FM = FIRM
    FRM = FIRM
    HD = HARD
    HRD = HARD
